/*
 * mp3cut: cuts out parts of an mp3 audio file
 * Copyright (C) 2004, 2005 Christian Wolff
 * 2005-03-14
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 * 
 * The author can be reached at sub-mp3cut<at>scara.com, 
 * the project's page is at http://scara.com/mp3cut/
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifndef PATH_MAX
#define PATH_MAX 2048
#endif

#define def_bitrate 128000
#define def_cut_start 0  // cut from
#define def_cut_end 100   // cut to

// Search for next occurence of 0xFFFB in 'file' at or after 'start'
static long find_mp3_header(FILE *file, long start)
{
	unsigned char ch;
//	unsigned char match[2] = {0xFF, 0xFB};
	int n = 0;
	
	if (fseek(file, start, SEEK_SET) < 0) {
		fprintf(stderr, "Failed to seek to %lu!\n", start);
		return -1;
	}
	do {
		if (fread(&ch, 1, 1, file) < 1) break;
		if (n == 0) {
			if (ch == 0xFF) n++;
			else n = 0;
		} else {
			if (((ch & 0xF0) == 0xF0) && (ch & 0x06)) n++;
			else if (ch == 0xFF) n = 1;
			else n = 0;
		}
//		if (ch == match[n]) n++;          // match next?
//		else if (ch == match[0]) n = 1;  // match first?
//		else n = 0;                     // no match
//		if (n == sizeof(match)) 
		if (n > 1) 
			return ftell(file) - n;
	} while (1);
	return -1;
}

// Copy 'size' bytes from open file 'in' to open file 'out'
// returns number of bytes copied, or -1
static long fcopy(FILE *in, FILE *out, long size)
{
	unsigned char buffer[64 * 1024];
	long block, chunk, offs, copied = 0;
	
	block = sizeof(buffer);
	do {
		offs = 0;
		if (block > (size - copied)) block = size - copied;
		do {
			chunk = fread(&buffer[offs], 1, block - offs, in);
			if (chunk <= 0) break;
			offs += chunk;
		} while (offs < block);
		if (offs < block) block = offs;
		offs = 0;
		do {
			chunk = fwrite(&buffer[offs], 1, block - offs, out);
			if (chunk <= 0) break;
			offs += chunk;
		} while (offs < block);
		copied += offs;
		if (offs < sizeof(buffer)) break;
	} while (copied < size);
	return copied;
}

int main(int argc, char *argv[])
{
	int arg_in, arg_out, c;
	extern char *optarg;
	extern int optind, opterr, optopt;
	char *endptr;
	unsigned long param;
	unsigned long bitrate, cut_start, cut_end;
	int id3remove, truncate;
	FILE *in, *out;
	long start, end, filestart, fileend, size, copied;
	char target[PATH_MAX];
	
	printf("mp3cut 0.8 (c) 2003-2006 Christian Wolff\n");
	// Get parameter
	bitrate = def_bitrate;
	cut_start = def_cut_start;
	cut_end = def_cut_end;
	id3remove = 0;
	truncate = 0;
	while (1) {
		c = getopt(argc, argv, "b:s:e:it");
		if (c < 0) break;
		switch (c) {
			case 'b':
				param = strtoul(optarg, &endptr, 0);
				if (endptr[0] == '\0') bitrate = param * 1000;
				break;
			case 's':
				param = strtoul(optarg, &endptr, 0);
				if (endptr[0] == '\0') cut_start = param;
				break;
			case 'e':
				param = strtoul(optarg, &endptr, 0);
				if (endptr[0] == '\0') cut_end = param;
				break;
			case 'i':
				id3remove = 1;
				break;
			case 't':
				truncate = 1;
				break;
			default:
				optind = argc;
		}
	}
	if (optind < argc) arg_in = optind++;
	else {
		printf("Removes a section from an MPEG1 Layer III audio file. (c) 2003 Christian Wolff.\n");
		printf("usage: %s <input file> [<output file>] [-b <bitrate>] [-s <cut start>] [-e <cut end after start>] [-i] [-t]\n", argv[0]);
		printf("  -b: specify bitrate of the mp3 file, in kBit/s (1000 bit per second), default: %ld\n", def_bitrate);
		printf("  -s: start of section to be cut out, in tenth of a second, default: %ld\n", def_cut_start);
		printf("  -e: end of section to cut out, in 1/10th sec. after start (actually the length of the cut out section), default: %ld\n", def_cut_end);
		printf("  -i: remove ID3 section at beginning of the file\n");
		printf("  -t: truncate the last mp3 block from the end, to clean the end of the file\n");
		exit(0);
	}
	if (optind < argc) arg_out = optind++;
	else arg_out = 0;
	
	// Print status
	printf("Source file: %s\n", argv[arg_in]);
	if (arg_out) printf("Target file: %s\n", argv[arg_out]);
	printf("Bitrate: %d kBit/s, cut from %d.%1d to %d.%1d s\n", 
		bitrate / 1000, 
		cut_start / 10, cut_start % 10, 
		(cut_start + cut_end) / 10, (cut_start + cut_end) % 10);
	
	// Open source file
	in = fopen(argv[arg_in], "rb");
	if (in == NULL) {
		fprintf(stderr, "File not found: %s\n", argv[arg_in]);
		exit(1);
	}
	
	// Find start and end, aligned to mp3 block
	filestart = find_mp3_header(in, find_mp3_header(in, 0) + 4);
	start = find_mp3_header(in, filestart + (long)((long long)cut_start * (bitrate >> 3) / 10));
	if (start < 0) {
		fprintf(stderr, "Failed to find MP3 header %d.%1d seconds into the file\n", 
			cut_start / 10, cut_start % 10);
		exit(1);
	}
	end = find_mp3_header(in, start + (long)((long long)cut_end * (bitrate >> 3) / 10));
	if (end < 0) {
		fprintf(stderr, "Failed to find MP3 header %d.%1d seconds into the file, deleting until end of file\n", 
			(cut_start + cut_end) / 10, (cut_start + cut_end) % 10);
		fseek(in, 0, SEEK_END);
		end = ftell(in);
	}
	printf("Cutting from 0x%X to 0x%X,\n", start, end);
	if (! id3remove) filestart = 0;
	
	// Open destination file
	if (arg_out) {
		if (! strncmp(argv[arg_out], argv[arg_in], PATH_MAX)) {
			fprintf(stderr, "Input and Output can't be the same file!\n");
			exit(1);
		}
		printf("Writing to %s\n", argv[arg_out]);
		out = fopen(argv[arg_out], "wb");
		if (out == NULL) fprintf(stderr, "Error writing file: %s\n", argv[arg_out]);
	} else {
		strncpy(target, argv[arg_in], PATH_MAX);
		strncat(target, "_edit.mp3", PATH_MAX);
		printf("Writing to %s\n", target);
		out = fopen(target, "wb");
		if (out == NULL) fprintf(stderr, "Error writing file: %s\n", target);
	}
	if (out == NULL) exit(1);
	
	// Get source file size
	if (fseek(in, 0, SEEK_END) < 0) {
		fprintf(stderr, "Failed to seek to end of input file!\n");
		exit(1);
	}
	fileend = ftell(in);
	
	if (truncate) {
		unsigned long trs = fileend;
		do {
			trs -= 400;  // ballpark figure, needs to be calculated: 144 * bitrate / samplerate + padding
			fileend = find_mp3_header(in, trs);
		} while (fileend < 0);
	}
	
	// Copy start of file
	printf("Copying start of file (%ld bytes)\n", start - filestart);
	//rewind(in);
	if (fseek(in, filestart, SEEK_SET) < 0) {
		fprintf(stderr, "Failed to seek to %lu!\n", filestart);
		return -1;
	}
	copied = fcopy(in, out, start - filestart);
	if (copied < start - filestart) {
		fprintf(stderr, "Failed to copy %ld bytes, wrote %ld bytes instead!\n", start - filestart, copied);
		exit(1);
	}
	
	// Copy end of file
	printf("Copying end of file (%ld - %ld = %ld bytes)\n", 
		fileend, end, fileend - end);
	if (fseek(in, end, SEEK_SET) < 0) {
		fprintf(stderr, "Failed to seek to %ld in input file!\n", end);
		exit(1);
	}
	copied = fcopy(in, out, fileend - end);
	if (copied < fileend - end) {
		fprintf(stderr, "Failed to copy %ld bytes, wrote %ld bytes instead!\n", 
			fileend - end, copied);
		exit(1);
	}
	
	// Close files
	fclose(in);
	fclose(out);
	
	return 0;
}

